from sqlalchemy import Column, String, ForeignKey, Float, Numeric
from xplordb.datamodel.lith import Lith

from xplordb.sqlalchemy.base import Base


class XplordbLithTable(Lith, Base):
    """
    Define Xplordb columns for lith definition
    """
    __tablename__ = "lith"
    __table_args__ = {"schema": 'dh'}

    data_set = Column("data_set", String, ForeignKey('ref.data_sets.data_set'))
    hole_id = Column("hole_id", String, ForeignKey('dh.collar.hole_id'), primary_key=True)

    lith_code = Column('lith_code_1', String, ForeignKey('ref.lithology.code'))

    # Columns defined as Numeric so they can be used as primary keys
    # Values in Lith class defined as float : not used as decimal
    from_m = Column('from_m', Numeric(asdecimal=False), primary_key=True)
    to_m = Column('to_m', Numeric(asdecimal=False), primary_key=True)

    loaded_by = Column(String, ForeignKey('ref.person.code'))
    logged_by = Column(String, ForeignKey('ref.person.code'))



from sqlalchemy import Column, String, ForeignKey
from sqlalchemy.orm import relationship

from xplordb.datamodel.dataset import Dataset
from xplordb.sqlalchemy.base import Base


class XplordbDatasetTable(Dataset, Base):
    """
    Define spatialite columns for Dataset definition
    """
    __tablename__ = "data_sets"
    __table_args__ = {"schema": 'ref'}

    name = Column('data_set', String, primary_key=True)
    full_name = Column(String)
    loaded_by = Column(String, ForeignKey('ref.person.code'))

    collars = relationship("XplordbCollarTable")
    surveys = relationship("XplordbSurveyTable")
    liths = relationship("XplordbLithTable")





"""
Short description of this Python module.

Longer description of this module.

xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/02"
__license__ = "AGPLv3"

import math
import unittest
from typing import List

import pytest
import sqlalchemy
from pytest import approx
from sqlalchemy import text

from xplordb.minimal_dao import MinimalDao

PERSON = 'xdb'  # Maximum 3 char
DATA_SET = 'test'
COLLAR = 'collar'

X_COLLAR = 100
Y_COLLAR = 150
Z_COLLAR = 0
SRID = 3857
EOH = 775.0


def test_person_import_lite_db(db_session, litedb_with_data):
    """
    Test if person sample data are imported with xplordb lite
    """
    _check_person_table_import(db_session, 6)


def test_person_import_full_db(db_session, fulldb_with_data):
    """
    Test if person sample data are imported with xplordb full
    """
    _check_person_table_import(db_session, 6)


def test_person_import_no_data(db_session, litedb_no_data):
    """
    Test that no person are imported with xplordb lite without sample data
    """
    _check_person_table_import(db_session, 0)


def test_person_import_no_data_full_db(db_session, fulldb_no_data):
    """
    Test that no person are imported with xplordb lite without sample data
    """
    _check_person_table_import(db_session, 0)


def _check_person_table_import(db_session, expected_number: int):
    """
    Test number of person available in xplordb

    :param db_session: psycog2 connection
    :param expected_number: expected number of person
    """
    with db_session.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM ref.person")
        rec = cur.fetchone()
        assert expected_number == rec[0]


def test_person_creation(sqlalchemy_session, litedb_no_data):
    """
    Test person creation with xplordb lite
    """
    dao = MinimalDao(sqlalchemy_session)
    dao.create_person(PERSON)
    sqlalchemy_session.commit()

    rec = sqlalchemy_session.execute(text("SELECT code FROM ref.person")).first()
    assert PERSON == rec[0]


def test_collar_creation(sqlalchemy_session, litedb_no_data):
    """
    Test collar creation without survey on xplordb lite
    """
    dao = MinimalDao(sqlalchemy_session)
    dao.create_collar(COLLAR, DATA_SET, PERSON)
    sqlalchemy_session.commit()

    rec = sqlalchemy_session.execute("SELECT hole_id,geom_trace FROM dh.collar").first()
    assert COLLAR == rec[0]
    assert not rec[1]


def test_collar_surv_creation(sqlalchemy_session, litedb_no_data):
    """
    Test collar creation with survey on xplordb lite
    """
    _create_and_check_collar_surv_creation(sqlalchemy_session)


def test_collar_surv_creation_full_database(sqlalchemy_session, fulldb_no_data):
    """
    Test collar creation with survey on xplordb full
    """
    _create_and_check_collar_surv_creation(sqlalchemy_session)


def _create_and_check_collar_surv_creation(sqlalchemy_session):
    """
    Create a collar and a simple survey and check that geometry is created

    :param db_session: psycog2 connection
    """
    surveys = _simple_collar_surv_creation(sqlalchemy_session, PERSON, DATA_SET, COLLAR, X_COLLAR, Y_COLLAR,
                                           Z_COLLAR,
                                           SRID, EOH)

    depth, x, y, z = _minimum_curvature_method(surveys, X_COLLAR, Y_COLLAR, Z_COLLAR, EOH)
    _check_table_geom_trace(sqlalchemy_session, x, y, z, depth, SRID, 'collar')


def _simple_collar_surv_creation(sqlalchemy_session, person: str, data_set: str,  collar: str,
                                 x_collar: float, y_collar: float, z_collar: float, srid: int, eoh: float = None):
    """
    Create a collar and a simple survey

    :param db_session: psycog2 connection
    :param person: person associated with collar
    :param data_set: data set associated with collar
    :param collar: collar name unique identifier
    :param x_collar: collar x coordinate
    :param y_collar: collar y coordinate
    :param z_collar: collar z coordinate
    :param srid: coordinate srid
    :param eoh: (optional) collar end of hole
    :return: list of surveys used
    """
    dao = MinimalDao(sqlalchemy_session)
    dao.create_collar(collar, data_set, person, x_collar, y_collar, z_collar, srid, eoh)

    surveys = [{'depth': 0.0, 'dip': 45.0, 'azimuth': 0.0},
               {'depth': 100.00, 'dip': 45.0, 'azimuth': 0.0},
               {'depth': 200.00, 'dip': 45.0, 'azimuth': 5.0},
               {'depth': 300.00, 'dip': 45.0, 'azimuth': 7.0},
               {'depth': 400.00, 'dip': 45.0, 'azimuth': 6.0}, ]
    for survey in surveys:
        dao.create_surv(survey['depth'], survey['dip'], survey['azimuth'], collar, data_set, person)

    sqlalchemy_session.commit()

    return surveys


def _check_table_geom_trace(sqlalchemy_session, x: List[float], y: List[float], z: List[float], depth: List[float],
                            srid: int, dh_table: str):
    """
    Check table geom_trace geometry from references values

    :param db_session: psycog2 connection
    :param x: expected point list x coordinates
    :param y: expected point list y coordinates
    :param z: expected point list z coordinates
    :param depth: expected point list m coordinates
    :param srid: point srid
    :param dh_table: table to check
    """
    values = sqlalchemy_session.execute(text(
        f"with point_list AS (SELECT (ST_DumpPoints(ST_Transform(geom_trace,{srid}))).geom as geom FROM dh.{dh_table})"
        "SELECT ST_X(geom), ST_Y(geom), ST_Z(geom),ST_M(geom) FROM point_list ORDER BY ST_Z(geom)")).all()
    assert len(values) == len(x)
    for i in range(0, len(values)):
        assert values[i][0] == approx(x[i])
        assert values[i][1] == approx(y[i])
        assert values[i][2] == approx(z[i])
        assert values[i][3] == approx(depth[i])


def _minimum_curvature_method(surveys, x_collar: float, y_collar: float, z_collar: float, eoh: float):
    """
    Define point list coordinates from Minimum curvature method implementation from https://www.drillingformulas.com/minimum-curvature-method/

    :param surveys: surveys list
    :param x_collar: collar x coordinate
    :param y_collar: collar y coordinate
    :param z_collar: collar z coordinate
    :return: depth, x, y, z : list of calculated coordinates
    """

    # A survey at depth 0.0 if not available
    if len(surveys):
        if surveys[0]['depth'] != 0.0:
            surveys.insert(0,{"dip": surveys[0]['dip'],
                            "azimuth": surveys[0]['azimuth'],
                            "depth": 0
                            }
                           )
    # append eoh survey
    surveys.append({'depth': eoh, 'dip': 45.0, 'azimuth': 6.0})

    surveys.sort(key = lambda x : x["depth"])

    depth_collar = 0.0
    x = [x_collar]
    y = [y_collar]
    z = [z_collar]
    depth = [depth_collar]
    current_x = x_collar
    current_y = y_collar
    current_z = z_collar
    current_depth = depth_collar
    for i in range(0, len(surveys) - 1):
        s2 = surveys[i + 1]
        s1 = surveys[i]
        # In xplordb dip is changed (for convention ?)
        i2 = math.radians(s2['dip'] + 90.0)
        i1 = math.radians(s1['dip'] + 90.0)
        a2 = math.radians(s2['azimuth'])
        a1 = math.radians(s1['azimuth'])
        CL = s2['depth'] - s1['depth']
        DL = math.acos(math.cos(i2 - i1) - (math.sin(i1) * math.sin(i2)) * (1 - math.cos(a2 - a1)))

        # Formula with different https//directionaldrillingart.blogspot.com/2015/09/directional-surveying-calculations.html
        # Produce same results
        # DL = math.acos((math.sin(i1) * math.sin(i2) * math.cos(a2 - a1)) + (math.cos(i1) * math.cos(i2)))
        RF = 1.0
        if DL != 0.0:
            # Formula with different method https//directionaldrillingart.blogspot.com/2015/09/directional-surveying-calculations.html
            # RF = math.tan(DL / 2) * (180 / math.pi) * (2 / DL)
            RF = math.tan(DL / 2) * (2 / DL)

        current_x = ((math.sin(i1) * math.sin(a1)) + (math.sin(i2) * math.sin(a2))) * (RF * (CL / 2)) + current_x
        current_y = ((math.sin(i1) * math.cos(a1)) + (math.sin(i2) * math.cos(a2))) * (RF * (CL / 2)) + current_y
        current_z = - (math.cos(i1) + math.cos(i2)) * (CL / 2) * RF + current_z
        current_depth += CL
        x.append(current_x)
        y.append(current_y)
        z.append(current_z)
        depth.append(current_depth)
    return depth, x, y, z


#def test_collar_details_surv_creation(sqlalchemy_session, litedb_no_data):
#    """
#    Test collar creation with survey and details on xplordb lite
#    """
#    _create_and_check_collar_surv_details_creation(sqlalchemy_session)


#def test_collar_details_surv_creation_full_database(sqlalchemy_session, fulldb_no_data):
#    """
#    Test collar creation with survey and details on xplordb full
#    """
#    _create_and_check_collar_surv_details_creation(sqlalchemy_session)


def _create_and_check_collar_surv_details_creation(sqlalchemy_session):
    """
    Create a collar with a simple survey and details and check that geometry is created

    :param db_session: psycog2 connection
    """
    surveys = _collar_with_details_creation(sqlalchemy_session, PERSON, DATA_SET, COLLAR, X_COLLAR, Y_COLLAR,
                                            Z_COLLAR, SRID, EOH)
    depth, x, y, z = _minimum_curvature_method(surveys, X_COLLAR, Y_COLLAR, Z_COLLAR, EOH)
    _check_table_geom_trace(sqlalchemy_session, x, y, z, depth, SRID, 'collar')
    _check_table_geom_trace(sqlalchemy_session, x, y, z, depth, SRID, 'details')

    # Add new survey and check that details geom_trace is updated with new survey
    dao = MinimalDao(sqlalchemy_session)
    new_surveys = [{'depth': 750.0, 'dip': 45.0, 'azimuth': 8.0}, ]
    for survey in new_surveys:
        dao.create_surv(survey['depth'], survey['dip'], survey['azimuth'], COLLAR, DATA_SET, PERSON)
        surveys.append(survey)
    sqlalchemy_session.commit()

    depth, x, y, z = _minimum_curvature_method(surveys, X_COLLAR, Y_COLLAR, Z_COLLAR, EOH)
    _check_table_geom_trace(sqlalchemy_session, x, y, z, depth, SRID, 'collar')
    _check_table_geom_trace(sqlalchemy_session, x, y, z, depth, SRID, 'details')


def _collar_with_details_creation(sqlalchemy_session, person: str, data_set: str, collar: str,
                                  x_collar: float, y_collar: float, z_collar: float, srid: int, eoh: float):
    """
    Create a collar with simple survey and add details

    :param db_session: psycog2 connection
    :param person: person associated with collar
    :param data_set: data set associated with collar
    :param collar: collar name unique identifier
    :param x_collar: collar x coordinate
    :param y_collar: collar y coordinate
    :param z_collar: collar z coordinate
    :param srid: coordinate srid
    :return: list of surveys used
    """
    surveys = _simple_collar_surv_creation(sqlalchemy_session, person, data_set, collar,
                                           x_collar, y_collar, z_collar, srid, eoh)
    dao = MinimalDao(sqlalchemy_session)
    dao.create_details(0.0, 750.0, collar, data_set, person)
    sqlalchemy_session.commit()
    return surveys


def test_surv_depth_check_with_eoh(sqlalchemy_session, litedb_no_data):
    """
    Test survey creation outside collar details
    """
    _simple_collar_surv_creation(sqlalchemy_session, PERSON, DATA_SET, COLLAR, X_COLLAR, Y_COLLAR, Z_COLLAR, SRID,
                                 750.0)
    # Add new survey after details maximum depth
    expected_message = "error! depth exceeds maximum hole depth (750). Check input or add details for collar in dh.details."
    with pytest.raises(sqlalchemy.exc.InternalError) as expect:
        dao = MinimalDao(sqlalchemy_session)
        new_surveys = [{'depth': 800.0, 'dip': 45.0, 'azimuth': 8.0}, ]
        for survey in new_surveys:
            dao.create_surv(survey['depth'], survey['dip'], survey['azimuth'], COLLAR, DATA_SET, PERSON)
        sqlalchemy_session.commit()
    assert expected_message in str(expect.value)


def test_trace_update_tables_all_function(sqlalchemy_session, litedb_no_data):
    """
    Test call for function dh.trace_update_tables_all() on xplordb lite

    For now only check if no error is raised when executing function
    """
    _check_trace_update_tables_all_function(sqlalchemy_session)


def test_trace_update_tables_all_function_full_database(sqlalchemy_session, fulldb_no_data):
    """
    Test call for function dh.trace_update_tables_all() on xplordb full

    For now only check if no error is raised when executing function
    """
    _check_trace_update_tables_all_function(sqlalchemy_session)


def _check_trace_update_tables_all_function(sqlalchemy_session):
    """
    Create a collar with survey and details and check dh.trace_update_tables_all() execution

    :param db_session: psycog2 connection
    """
    _collar_with_details_creation(sqlalchemy_session, PERSON, DATA_SET, COLLAR, X_COLLAR, Y_COLLAR, Z_COLLAR, SRID, EOH)
    sqlalchemy_session.execute(text("SELECT dh.trace_update_tables_all()"))


def test_trace_update_tables_hole_function_full_database(sqlalchemy_session, fulldb_no_data):
    """
    Test call for function dh.trace_update_tables_hole() on xplordb full

    For now only check if no error is raised when executing function
    """
    _check_trace_update_tables_hole(sqlalchemy_session)


def test_trace_update_tables_hole_function(sqlalchemy_session, litedb_no_data):
    """
    Test call for function dh.trace_update_tables_hole() on xplordb lite

    For now only check if no error is raised when executing function
    """
    _check_trace_update_tables_hole(sqlalchemy_session)


def _check_trace_update_tables_hole(sqlalchemy_session):
    """
    Create a collar with survey and details and check dh.trace_update_tables_hole() execution

    :param db_session: psycog2 connection
    """
    _collar_with_details_creation(sqlalchemy_session, PERSON, DATA_SET, COLLAR, X_COLLAR, Y_COLLAR, Z_COLLAR, SRID, EOH)
    sqlalchemy_session.execute(text(f"SELECT dh.trace_update_tables_hole('{COLLAR}')"))


def test_survey_incomplete_azimuth_values(sqlalchemy_session, litedb_no_data):
    """
    Test exception raised if survey with azimuth defined an no azimuth date
    """
    dao = MinimalDao(sqlalchemy_session)
    dao.create_collar(COLLAR, DATA_SET, PERSON, X_COLLAR, Y_COLLAR, Z_COLLAR, SRID)
    expected_message = 'new row for relation "surv" violates check constraint "dh_surv_check_azimuth_date"'
    with pytest.raises(sqlalchemy.exc.IntegrityError) as expect:
        dao.create_invalid_surv(0.0, 0.0, 0.0, COLLAR, DATA_SET, PERSON)
    assert expected_message in str(expect.value)


if __name__ == '__main__':
    unittest.main()



"""
Module use to import xplordb DDL in an existing database

Module use to import xplordb DDL in an existing database.
Options are available to import full database and example data

xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/02"
__license__ = "AGPLv3"

import os
from pathlib import Path
from typing import List

from xplordb.schema import Schema
from xplordb.sql_utils import connect


class ImportDDL:
    IMPORT_ROOT = Path(os.path.dirname(os.path.abspath(__file__)))

    def __init__(self, db: str, user: str, password: str, host: str = 'localhost', port: int = 5432):
        """
        A class used to import xplordb DDL to an existing database

        Example:
            import_ddl = ImportDDL(db='xplordb', user='postgres', password='postgres')
            import_ddl.import_xplordb_schema()

        :param db: database name where xplordb DDL will be imported
        :param user: user for database connection
        :param password: password for database connection
        :param host: host (optional, default 'localhost')
        :param port: port (optional, default 5432)
        """
        self._host = host
        self._port = port
        self._db = db
        self._user = user
        self._password = password

        self._conn = connect(
            database=db,
            host=host,
            port=port,
            user=user,
            password=password)

    def __del__(self):
        if self._conn:
            self._conn.close()

    @staticmethod
    def xplordb_roles():
        return ['xdb_viewer', 'xdb_logger', 'xdb_importer', 'xdb_admin']

    def import_xplordb_schema(self, import_data: bool = False, full_db: bool = False):
        """
        Import xplordb DDL to database
        :param import_data: boolean to define if sample dataset must be imported (optional, default False)
        :param full_db: boolean to define if full database DDL must be imported (optional, default False)
        """
        self._create_extension()

        self._create_roles()

        # Define schema list depending on full_db option
        schemas = Schema.db_schema_list(full_db)

        # Drop and import all schemas
        for schema in schemas:
            self._drop_and_import_schema(schema)

        # Import data if asked
        if import_data:
            self._import_schemas_data(schemas)

        # Import specific foreign keys and trigger AFTER data import because of some foreign keys and trigger
        for schema in schemas:
            schema.import_fk_creation(self._conn)
            schema.import_trigger(self._conn)

    def _drop_and_import_schema(self, schema: Schema):
        self._drop_schema(schema.name)
        self._import_schema(schema)

    def _drop_schema(self, schema: str):
        with self._conn.cursor() as cur:
            print(f'deleting schema {schema}')
            cur.execute(f'DROP SCHEMA IF EXISTS {schema} CASCADE')
            cur.close()
        self._conn.commit()

    def _import_schema(self, schema: Schema):
        schema.create_schema(self._conn)
        # Function are used in table definition and must be imported before table
        schema.import_pre_table_functions(self._conn)
        schema.import_tables(self._conn)
        # Still there are some function that refers to a table so they must be imported after table
        schema.import_post_table_functions(self._conn)
        schema.import_sequences(self._conn)
        schema.import_views(self._conn)

    def _import_schemas_data(self, schemas: List[Schema]):
        for schema in schemas:
            schema.import_tables_data(self._conn)

    def _create_extension(self):
        extensions = ['btree_gist', 'postgis', 'postgis_raster', 'plpgsql']
        with self._conn.cursor() as cur:
            for extension in extensions:
                cur.execute(f"CREATE EXTENSION IF NOT EXISTS {extension}")

    def _create_roles(self):
        for role in self.xplordb_roles():
            self._create_role(role)

    def _create_role(self, role : str):
        if not self._role_exist(role):
            with self._conn.cursor() as cur:
                cur.execute(f"CREATE ROLE {role}")

    def _role_exist(self, role: str):
        exist = False
        with self._conn.cursor() as cur:
            cur.execute(f"SELECT * FROM pg_roles WHERE rolname='{role}'")
            if cur.fetchone():
                exist = True
        return exist



from geoalchemy2 import Geometry
from sqlalchemy import Column, String, Float

from xplordb.datamodel.metadata import CollarMetadata, RawCollarMetadata
from xplordb.sqlalchemy.base import Base

DEFAULT_SRID = 4326



class XplordbCollarMetadataTable(CollarMetadata, Base):
    """
    Define Xplordb columns for Collar metadata definition
    """
    __tablename__ = "metadata"
    __table_args__ = {"schema": 'dh'}
    hole_id = Column("hole_id", String, primary_key = True)  
  

def create_xplordbcollarmetadataTable(collar_metadata:RawCollarMetadata) -> tuple:
    """
    Function returning updated XplordbCollarMetadataTable class instance for a specific hole_id and table definition
    """
    # we should derive base class to not alter it.
    class tmp(XplordbCollarMetadataTable):
        pass
    metadata_base = tmp

    for col_name, value in collar_metadata.extra_cols.items():
        data_type = String if type(value) == str else Float
        column = Column(col_name, data_type)
        if not hasattr(metadata_base, col_name):
            setattr(metadata_base, col_name, column)

    metadata_base_instance = metadata_base(collar_metadata.hole_id)
    for col_name, value in collar_metadata.extra_cols.items():
        
        setattr(metadata_base_instance, col_name, value)

    return metadata_base_instance, metadata_base


"""
Utility functions for postgresql database use

Utility functions for postgresql database use :
- psycog2 connection
- sql file import with psql (for data import)
- sql file import with psycog2 (for database definition)

xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/02"
__license__ = "AGPLv3"

import os
import subprocess
import chevron
from pathlib import Path
from typing import List

import psycopg2


class InvalidConnection(Exception):
    pass


def connect(**kwargs):
    """
    creates a connection to PostgreSQL and returns the instance
    """
    try:
        conn = psycopg2.connect(**kwargs)
    except psycopg2.Error as exc:
        raise InvalidConnection(exc)
    return conn


def exec_conn_psql(conn, filename: Path, mustache_params=None):
    """
    Execute psql code from a file with psycog2.
    File can contains mustache parameters such as {{params}}. They will be replace with mustache_params map contents
    :param conn: psycopg2 connection
    :param filename: Path to file
    :param mustache_params: (optional) map with mustache parameters
    """
    with conn.cursor() as cur:
        with open(filename, 'r') as f:
            cur.execute(chevron.render(f, mustache_params))
    conn.commit()


def import_all_sql_in_directory(conn, directory: Path, excluded_files: List[str] = None, mustache_params=None):
    """
    Execute psql code with psycog2 for all files inside a directory.
    File can contains mustache parameters such as {{params}}. They will be replace with mustache_params map contents
    :param conn: psycopg2 connection
    :param directory: directory parsed for files
    :param excluded_files: (optional) list of file that won't be imported
    :param mustache_params: (optional) map with mustache parameters
    """
    if directory.exists():
        for filename in os.listdir(directory):
            if not excluded_files or filename not in excluded_files:
                print(f'importing : {filename}')
                exec_conn_psql(conn, directory / filename, mustache_params)


def import_sql_in_directory(conn, directory: Path, files: List[str], mustache_params=None):
    """
    Execute psql code with psycog2 for specific files inside a directory.
    Warning : if file is not present in directory, an exception is raised
    File can contains mustache parameters such as {{params}}. They will be replace with mustache_params map contents
    :param conn: psycopg2 connection
    :param directory: directory parsed for files
    :param files: list of imported files
    :param mustache_params: (optional) map with mustache parameters
    """
    if files:
        for filename in files:
            print(f'importing : {filename}')
            exec_conn_psql(conn, directory / filename, mustache_params)


def exec_psql(conn, filename: Path):
    """
    Execute psql import with a file
    :param connn: psycopg2 connection used to get connection parameters
    :param filename: file to be imported
    """
    env = {'PGPASSWORD': conn.info.password}
    subprocess.run(['psql',
                    '-h', conn.info.host,
                    '-p', str(conn.info.port),
                    '-U', conn.info.user,
                    '--dbname', conn.info.dbname,
                    '-f', str(filename)],
                   env=env)


def psql_all_sql_in_directory(conn, directory: Path, excluded_files: List[str] = None):
    """
    Execute psql import for all files inside a directory.
    :param conn: psycopg2 connection used to get connection parameters
    :param directory: directory parsed for files
    :param excluded_files: (optional) list of file that won't be imported
    """
    if directory.exists():
        for filename in os.listdir(directory):
            if not excluded_files or filename not in excluded_files:
                print(f'importing : {filename}')
                exec_psql(conn, directory / filename)


def psql_sql_in_directory(conn, directory: Path, files: List[str]):
    """
    Execute psql import for specific files inside a directory.
    Warning : if file is not present in directory, a exception is raised
    :param conn: psycopg2 connection used to get connection parameters
    :param directory: directory parsed for files
    :param files: list of imported files
    """
    if files:
        for filename in files:
            exec_psql(conn, directory / filename)



import json
from sqlalchemy.types import UserDefinedType

class pgAzimuthType(UserDefinedType):
    def get_col_spec(self):
        return "assay.azimuth"
    
class pgDipType(UserDefinedType):
    def get_col_spec(self):
        return "assay.dip"
    
class pgSphericalObjectType(UserDefinedType):
    def get_col_spec(self):
        return "assay.kind"
    
class pgPolarityType(UserDefinedType):
    def get_col_spec(self):
        return "assay.polarity"

class pgSphericalType(UserDefinedType):
    def get_col_spec(self):
        return "assay.spherical_data"
    
    @staticmethod
    def insert(azimuth: float, dip:float, polarity:int = 0, type: str = 'line'):
        res = f'({azimuth},{dip},{int(polarity)},{type.lower()})'
        return res

    @staticmethod
    def insert_from_dict(d: dict):
        return pgSphericalType.insert(**d)
    
    @staticmethod
    def insert_from_json(j: str):
        d = json.loads(j)
        return pgSphericalType.insert_from_dict(d)


from geoalchemy2 import Geometry
from sqlalchemy import Column, String, ForeignKey, Float, Integer, DateTime

from xplordb.datamodel.collar import Collar
from xplordb.sqlalchemy.base import Base

DEFAULT_SRID = 4326


class XplordbCollarTable(Collar, Base):
    """
    Define Xplordb columns for Collar definition
    """
    __tablename__ = "collar"
    __table_args__ = {"schema": 'dh'}

    hole_id = Column("hole_id", String, primary_key=True)
    data_set = Column("data_set", String, ForeignKey('ref.data_sets.data_set'))
    x = Column("x", Float)
    y = Column("y", Float)
    z = Column("z", Float)
    srid = Column("srid", Integer)
    project_srid = Column("project_srid", Integer)
    eoh = Column("eoh", Float)
    planned_eoh = Column("planned_eoh", Float)
    dip = Column("dip", Float)
    azimuth = Column("azimuth", Float)
    loaded_by = Column("loaded_by", String)
    survey_date = Column("survey_date", DateTime)
    geom = Column(Geometry(geometry_type='POINT', dimension=3, srid=DEFAULT_SRID))
    geom_trace = Column(Geometry(geometry_type='LINESTRING', dimension=3, srid=DEFAULT_SRID))
    planned_trace = Column(Geometry(geometry_type='LINESTRING', dimension=3, srid=DEFAULT_SRID))



from sqlalchemy import Column, String, ForeignKey, Float, Numeric

from xplordb.datamodel.survey import Survey

from xplordb.sqlalchemy.base import Base


class XplordbSurveyTable(Survey, Base):
    """
    Define Xplordb columns for Survey definition
    """
    __tablename__ = "surv"
    __table_args__ = {"schema": 'dh'}

    data_set = Column("data_set", String, ForeignKey('ref.data_sets.data_set'))
    hole_id = Column("hole_id", String, ForeignKey('dh.collar.hole_id'), primary_key=True)
    loaded_by = Column("loaded_by", String)

    # Column defined as Numeric so they can be used as primary keys
    # Value in Survey class defined as float : not used as decimal
    depth = Column("depth_m", Numeric(asdecimal=False), primary_key=True)
    dip = Column("dip", Float)
    azimuth = Column("azimuth_grid", Float)



from psycopg2 import connect
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

from xplordb.import_ddl import ImportDDL


class CreateDatabase:

    def __init__(self, connection_db: str, user: str, password: str, host: str = 'localhost',
                 port: int = 5432):
        """
        A class used to create xplordb database to an existing server

        Example:
            create_db = CreateDatabase(connection_db='postgres', user='postgres', password='postgres')
            create_db.create_db('xplordb', True, True)

        :param connection_db: database name for connection
        :param user: user for database connection
        :param password: password for database connection
        :param host: host (optional, default 'localhost')
        :param port: port (optional, default 5432)
        """
        self._host = host
        self._port = port
        self._user = user
        self._password = password

        self._conn = connect(
            database=connection_db,
            host=host,
            port=port,
            user=user,
            password=password)

    def __del__(self):
        if self._conn:
            self._conn.close()

    def create_db(self, db: str, import_data: bool = False, full_db: bool = False):
        """
        Create a new xplordb database

        :param db: new database name where xplordb DDL will be imported
        :param import_data: boolean to define if sample dataset must be imported (optional, default False)
        :param full_db: boolean to define if full database DDL must be imported (optional, default False)
        """
        # Create a table in PostgreSQL database
        self._conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cur = self._conn.cursor()
        cur.execute(f'CREATE DATABASE {db}')
        self._conn.commit()

        # Import DDL
        import_ddl = ImportDDL(db=db, user=self._user, password=self._password, host=self._host, port=self._port)
        import_ddl.import_xplordb_schema(import_data, full_db)



"""
Module use to import data into an existing xplordb database

Module use to import data into an existing xplordb database


xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/09"
__license__ = "AGPLv3"

import psycopg2
import sqlalchemy

from xplordb.datamodel.collar import Collar
from xplordb.datamodel.metadata import RawCollarMetadata, create_collar_metadata
from xplordb.sqlalchemy.dh.metadata import create_xplordbcollarmetadataTable
from xplordb.datamodel.dataset import Dataset
from xplordb.datamodel.lith import Lith, LithCode
from xplordb.datamodel.person import Person
from xplordb.datamodel.survey import Survey

from xplordb.sqlalchemy.dh.collar import XplordbCollarTable
from xplordb.sqlalchemy.dh.lith import XplordbLithTable
from xplordb.sqlalchemy.dh.survey import XplordbSurveyTable
from xplordb.sqlalchemy.ref.person import XplordbPersonTable
from xplordb.sqlalchemy.ref.dataset import XplordbDatasetTable
from xplordb.sqlalchemy.ref.lithcode import XplordbLithCodeTable


class ImportData:
    class ImportException(Exception):
        pass

    class InvalidSchema(psycopg2.Error):
        def __init__(self, table_name: str):
            super().__init__(f'Table "{table_name}" does not exist. Check if xplordb is installed.')

    def __init__(self, db_session):
        """
        A class used to import data into a xplordb database

        Example:
            import_data = ImportData(session)
            person = Person('xdb')
            import_data.import_persons_array([person])

        :param db_session: pyscog2 session
        """
        self._db_session = db_session

    def commit(self) -> None:
        """
        Commit current changes

        raises ImportException in case of error

        """

        try:
            self._db_session.commit()
        except sqlalchemy.exc.IntegrityError as exc:
            self._db_session.rollback()
            raise ImportData.ImportException(exc)

    def get_available_person_codes(self) -> [str]:
        """
        Get list af available person codes in xplordb

        Returns: list of available persons
        """
        persons = self._db_session.query(XplordbPersonTable.code).all()
        return [person.code for person in persons]

    def import_persons_array(self, persons: [Person]):
        """
        Import persons into xplordb.

        :param persons: Persons array
        """
        self._db_session.add_all([XplordbPersonTable(trigram=person.code, full_name=person.full_name,
                                                     active=person.active, type=person.type)
                                  for person in persons])

    def get_available_dataset_names(self) -> [str]:
        """
        Get list af available datasets in xplordb

        Returns: list of available dataset names
        """
        datasets = self._db_session.query(XplordbDatasetTable.name).all()
        return [dataset.name for dataset in datasets]

    def import_datasets_array(self, datasets: [Dataset]):
        """
        Import datasets into xplordb.

        :param datasets: Dataset array
        """
        self._db_session.add_all(
            [XplordbDatasetTable(name=dataset.name, full_name=dataset.full_name, loaded_by=dataset.loaded_by)
             for dataset in datasets])

    def import_collars_array(self, collars: [Collar]):
        """
        Import collars into xplordb.

        :param collars: Collar array
        """
        self._db_session.add_all(
            [XplordbCollarTable(hole_id=collar.hole_id, data_set=collar.data_set, loaded_by=collar.loaded_by,
                                x=collar.x, y=collar.y, z=collar.z, srid=collar.srid, project_srid = collar.project_srid, eoh=collar.eoh,
                                planned_eoh=collar.planned_eoh, dip = collar.dip, azimuth = collar.azimuth,
                                survey_date=collar.survey_date)
             for collar in collars])
        
    def import_metadatas_array(self, metadatas: [RawCollarMetadata]):
        """
        Import metadata in xplordb.  
        Table is altered first, then values are updated.  
        """

        # add column if dont exists in database

        extra_cols = {}
        for m in metadatas:
            for col_name, value in m.extra_cols.items():
                sql_type = (
                "FLOAT" if type(value) != str else "VARCHAR"
            )
                extra_cols[col_name] = sql_type

        q = """
            SELECT column_name
            FROM information_schema.columns
            WHERE table_schema = 'dh'
            AND table_name   = 'metadata';
            """
        col_base = [r[0] for r in self._db_session.execute(q).fetchall()]

        for col_name, col_type in extra_cols.items():
             
            if col_name not in col_base:
                q = f"ALTER TABLE dh.metadata ADD COLUMN {col_name} {col_type} ;"
                self._db_session.execute(q)
                self.commit()


        # update data
        for metadata in metadatas:
            # generate a updated XplordbCollarMetadataTable definitio with extra columns
            _, cls = create_xplordbcollarmetadataTable(metadata)
            obj = self._db_session.query(cls).get(metadata.hole_id)
            for col_name, value in metadata.extra_cols.items():
                setattr(obj, col_name, value)

                self._db_session.commit()


    def import_surveys_array(self, surveys: [Survey]):
        """
        Import surveys into xplordb.

        :param surveys: Survey array
        """
        self._db_session.add_all(
            [XplordbSurveyTable(hole_id=survey.hole_id, data_set=survey.data_set, loaded_by=survey.loaded_by,
                                depth=survey.depth, dip=survey.dip, azimuth=survey.azimuth)
             for survey in surveys])

    def import_liths_array(self, liths: [Lith]):
        """
        Import liths into xplordb.

        :param liths: lith array
        """

        # Insert lith_code into database if not available
        unavailable_lith_code_tuple = [(lith.lith_code, lith.loaded_by) for lith in liths
                                       if not self._lith_code_available(lith.lith_code)]
        # Remove duplicates
        unavailable_lith_code_tuple = set(unavailable_lith_code_tuple)
        unavailable_lith_code_array = [LithCode(lith[0], lith[1]) for lith in unavailable_lith_code_tuple]

        self.import_lithology_codes_array(unavailable_lith_code_array)
        self._db_session.flush()
        self._db_session.add_all(
            [XplordbLithTable(data_set=lith.data_set, hole_id=lith.hole_id, loaded_by=lith.loaded_by,
                              lith_code=lith.lith_code,
                              from_m=lith.from_m, to_m=lith.to_m,
                              logged_by=lith.logged_by)
             for lith in liths])

    def _lith_code_available(self, lith_code: str):
        try:
            val = self._db_session.execute(f"SELECT COUNT(code) FROM ref.lithology WHERE code='{lith_code}'").first()
            return val[0] != 0
        except psycopg2.errors.UndefinedTable:
            raise ImportData.InvalidSchema('ref.lithology')

    def import_lithology_codes_array(self, lith_codes: [LithCode]):
        """
        Import lithology codes into xplordb.

        :param lith_codes: Lithology code array
        """
        self._db_session.add_all(
            [XplordbLithCodeTable(code=lith_code.code, description=lith_code.description, loaded_by=lith_code.loaded_by)
             for lith_code in lith_codes])



from datetime import datetime

import pytest
from sqlalchemy import text

from xplordb.datamodel.collar import Collar
from xplordb.datamodel.metadata import RawCollarMetadata
from xplordb.datamodel.dataset import Dataset
from xplordb.datamodel.lith import Lith
from xplordb.datamodel.person import Person
from xplordb.datamodel.survey import Survey
from xplordb.import_data import ImportData

def test_metadata_import(data_importer, db_session, litedb_no_data):

    # first, add some collars
    persons = [Person('xdb')]
    data_importer.import_persons_array(persons)

    datasets = [Dataset('test', 'xdb'), ]
    data_importer.import_datasets_array(datasets)

    collars = [Collar('collar', 'test',  'xdb', 100.0, 100.0, 0.0, 3857, 1000),
               Collar('collar2', 'test',  'xdb', 100.0, 100.0, 0.0, 3857, None),
               Collar('collar3', 'test',  'xdb', 100.0, 100.0, 0.0, 3857, None, None,datetime.now())]
    data_importer.import_collars_array(collars)
    data_importer.commit()

    # import metadata
    metadatas = [RawCollarMetadata(hole_id="collar", extra_cols={"col_a" : 10}),
                 RawCollarMetadata(hole_id="collar3", extra_cols={"col_b" : 5})]
    
    data_importer.import_metadatas_array(metadatas)
    data_importer.commit()
    expected = [("collar", 10, None), 
                ("collar2", None, None), 
                ("collar3", None, 5)]
    
    with db_session.cursor() as cur:
        cur.execute("SELECT * FROM dh.metadata ORDER BY hole_id")
        values = cur.fetchall()
        assert len(values) == len(collars)
        for value, exp in zip(values, expected):
            assert value == exp




def test_person_import(data_importer, db_session, litedb_no_data):
    """
    Test person import
    """
    persons = [Person('a', 'a'),
               Person('b', 'b'),
               Person('c', 'c'), ]
    data_importer.import_persons_array(persons)
    data_importer.commit()

    with db_session.cursor() as cur:
        cur.execute("SELECT code,loaded_by,person,type,active FROM ref.person")
        values = cur.fetchall()
        assert len(values) == len(persons)
        for i in range(0, len(values)):
            assert values[i][0] == persons[i].code
            assert values[i][1] == persons[i].loaded_by
            assert values[i][2] == persons[i].full_name
            assert values[i][3] == persons[i].type
            assert values[i][4] == persons[i].active


def test_get_available_persons(data_importer, db_session, litedb_no_data):
    """
    Test person read from database
    """
    test_person_import(data_importer, db_session, litedb_no_data)
    assert data_importer.get_available_person_codes() == ['a', 'b', 'c']


def test_dataset_import_invalid_person(data_importer, litedb_no_data):
    """
    Test dataset import with invalid person
    """
    with pytest.raises(ImportData.ImportException):
        datasets = [Dataset('test', 'invalid')]
        data_importer.import_datasets_array(datasets)
        data_importer.commit()


def test_dataset_import(data_importer, db_session, litedb_no_data):
    """
    Test dataset import
    """
    persons = [Person('xdb')]
    data_importer.import_persons_array(persons)

    datasets = [Dataset('test', 'xdb', 'full_test_name'),
                Dataset('test2', 'xdb', 'full_test2_name'), ]
    data_importer.import_datasets_array(datasets)
    data_importer.commit()

    with db_session.cursor() as cur:
        cur.execute("SELECT data_set,full_name,loaded_by FROM ref.data_sets")
        values = cur.fetchall()
        assert len(values) == len(datasets)
        for i in range(0, len(values)):
            assert values[i][0] == datasets[i].name
            assert values[i][1] == datasets[i].full_name
            assert values[i][2] == datasets[i].loaded_by


def test_get_available_datasets(data_importer, db_session, litedb_no_data):
    """
    Test datasets read from database
    """
    test_dataset_import(data_importer, db_session, litedb_no_data)
    assert data_importer.get_available_dataset_names() == ['test', 'test2']


def test_collar_import_invalid_fk(data_importer, litedb_no_data):
    """
    Test collar import with an invalid foreign key
    """
    with pytest.raises(ImportData.ImportException):
        collar = [Collar('collar', 'test', 'xdb', 100.0, 100.0, 0.0, 3857), ]
        data_importer.import_collars_array(collar)
        data_importer.commit()


def test_collar_import(data_importer, db_session, litedb_no_data):
    """
    Test collar import
    """
    persons = [Person('xdb')]
    data_importer.import_persons_array(persons)

    datasets = [Dataset('test', 'xdb'), ]
    data_importer.import_datasets_array(datasets)

    collars = [Collar('collar', 'test',  'xdb', 100.0, 100.0, 0.0, 3857, 1000),
               Collar('collar2', 'test',  'xdb', 100.0, 100.0, 0.0, 3857, None),
               Collar('collar3', 'test',  'xdb', 100.0, 100.0, 0.0, 3857, None, None,datetime.now())]
    data_importer.import_collars_array(collars)
    data_importer.commit()

    with db_session.cursor() as cur:
        cur.execute("SELECT data_set,hole_id,loaded_by,x,y,z,srid,eoh, planned_eoh,survey_date FROM dh.collar")
        values = cur.fetchall()
        assert len(values) == len(collars)
        for i in range(0, len(values)):
            assert values[i][0] == collars[i].data_set
            assert values[i][1] == collars[i].hole_id
            assert values[i][2] == collars[i].loaded_by
            assert values[i][3] == collars[i].x
            assert values[i][4] == collars[i].y
            assert values[i][5] == collars[i].z
            assert values[i][6] == collars[i].srid
            assert values[i][7] == collars[i].eoh
            assert values[i][8] == collars[i].planned_eoh
            if collars[i].survey_date:
                # Check date and time for no timezone comparison (not setted in now())
                assert values[i][9].date() == collars[i].survey_date.date()
                assert values[i][9].time() == collars[i].survey_date.time()


def test_survey_import_invalid_fk(data_importer, litedb_no_data):
    """
    Test survey import with an invalid foreign key
    """
    with pytest.raises(ImportData.ImportException):
        surveys = [Survey('collar', 'test', 'xdb', 0.0, 45.0, 0.0), ]
        data_importer.import_surveys_array(surveys)
        data_importer.commit()


def test_survey_import(data_importer, db_session, litedb_no_data):
    """
    Test survey import
    """
    persons = [Person('xdb')]
    data_importer.import_persons_array(persons)

    datasets = [Dataset('test', 'xdb'), ]
    data_importer.import_datasets_array(datasets)

    collars = [Collar('collar', 'test', 'xdb', 100.0, 100.0, 0.0, 3857), ]
    data_importer.import_collars_array(collars)

    surveys = [Survey('collar', 'test', 'xdb', 0.0, 45.0, 0.0), ]
    data_importer.import_surveys_array(surveys)
    data_importer.commit()

    with db_session.cursor() as cur:
        cur.execute("SELECT data_set,hole_id,loaded_by,depth_m,dip,azimuth_grid FROM dh.surv")
        values = cur.fetchall()
        assert len(values) == len(surveys)
        for i in range(0, len(values)):
            assert values[i][0] == surveys[i].data_set
            assert values[i][1] == surveys[i].hole_id
            assert values[i][2] == surveys[i].loaded_by
            assert values[i][3] == surveys[i].depth
            assert values[i][4] == surveys[i].dip
            assert values[i][5] == surveys[i].azimuth


def test_lith_import(data_importer, db_session, litedb_no_data):
    """
    Test lith import
    """
    persons = [Person('xdb')]
    data_importer.import_persons_array(persons)

    datasets = [Dataset('test', 'xdb'), ]
    data_importer.import_datasets_array(datasets)

    collars = [Collar('collar', 'test',  'xdb', 100.0, 100.0, 0.0, 3857), ]
    data_importer.import_collars_array(collars)

    liths = [Lith('lith1', 'test', 'collar', 'xdb',  0.0, 10.0),
             Lith('lith1', 'test', 'collar', 'xdb',  10., 100.0)]
    data_importer.import_liths_array(liths)
    data_importer.commit()

    with db_session.cursor() as cur:
        cur.execute("SELECT data_set,hole_id,loaded_by,lith_code_1,from_m,to_m FROM dh.lith")
        values = cur.fetchall()
        assert len(values) == len(liths)
        for i in range(0, len(values)):
            assert values[i][0] == liths[i].data_set
            assert values[i][1] == liths[i].hole_id
            assert values[i][2] == liths[i].loaded_by
            assert values[i][3] == liths[i].lith_code
            assert values[i][4] == liths[i].from_m
            assert values[i][5] == liths[i].to_m
